/*---------------------------------------------------------------------------*
 *                                   IT++			             *
 *---------------------------------------------------------------------------*
 * Copyright (c) 1995-2003 by Tony Ottosson, Thomas Eriksson, Pl Frenger,   *
 * Tobias Ringstrm, and Jonas Samuelsson.                                   *
 *                                                                           *
 * Permission to use, copy, modify, and distribute this software and its     *
 * documentation under the terms of the GNU General Public License is hereby *
 * granted. No representations are made about the suitability of this        *
 * software for any purpose. It is provided "as is" without expressed or     *
 * implied warranty. See the GNU General Public License for more details.    *
 *---------------------------------------------------------------------------*/

/*!
  \file
  \brief Implementation of functions on vectors and matrices.
  \author Tony Ottosson

  1.21

  2003/05/22 08:55:19
*/

#include "base/matfunc.h"

namespace itpp {

  double mean(const vec &v)
  {
    return sum(v)/v.length();
  }

  complex<double> mean(const cvec &v)
  {
    return sum(v)/static_cast<complex<double> >(v.length());
  }

  double mean(const svec &v)
  {
    return (double)sum(v)/v.length();
  }

  double mean(const ivec &v)
  {
    return (double)sum(v)/v.length();
  }

  template<>
  double norm(const cvec &v)
  {
    int i;
    double E=0.0;

    for (i=0; i<v.length(); i++)
      E += std::norm(v[i]);

    return sqrt(E);
  }

  template<>
  double norm(const cvec &v, int p)
  {
    int i;
    double E=0;

    for (i=0;i<v.length();i++)
      E+=pow(std::norm(v[i]), p/2.0); // Yes, 2.0 is correct!

    return pow(E,1.0/(double)p);
  }

  template<>
  double variance(const cvec &v)
  {
    int len = v.length();
    double sq_sum=0.0;
    complex<double> sum=0.0;
    const complex<double> *p=v._data();

    for (int i=0; i<len; i++, p++) {
      sum += *p;
      sq_sum += std::norm(*p);
    }

    return (double)(sq_sum - std::norm(sum)/len) / (len-1);
  }

  double mean(const mat &m)
  {
    return sum(m)/(m.rows()*m.cols());
  }

  complex<double> mean(const cmat &m)
  {
    return sum(m)/static_cast<complex<double> >(m.rows()*m.cols());
  }

  double mean(const smat &m)
  {
    return static_cast<double>(sum(m))/(m.rows()*m.cols());
  }

  double mean(const imat &m)
  {
    return static_cast<double>(sum(m))/(m.rows()*m.cols());
  }

  // ---------------------- Instantiations -----------------------------------------

  //! Template instantiation of help function QS for vec
  template void QS(int low, int high, vec &data);
  //! Template instantiation of help function QS for svec
  template void QS(int low, int high, svec &data);
  //! Template instantiation of help function QS for ivec
  template void QS(int low, int high, ivec &data);
  //! Template instantiation of help function QS for bvec
  template void QS(int low, int high, bvec &data);

  //! Template instantiation of help function QSindex for vec
  template void QSindex(int low, int high, ivec &indexlist, const vec &data);
  //! Template instantiation of help function QSindex for svec
  template void QSindex(int low, int high, ivec &indexlist, const svec &data);
  //! Template instantiation of help function QSindex for ivec
  template void QSindex(int low, int high, ivec &indexlist, const ivec &data);
  //! Template instantiation of help function QSindex for bvec
  template void QSindex(int low, int high, ivec &indexlist, const bvec &data);

  //! Template instantiation of length
  template int length(const vec &v);
  //! Template instantiation of length
  template int length(const cvec &v);
  //! Template instantiation of length
  template int length(const svec &v);
  //! Template instantiation of length
  template int length(const ivec &v);
  //! Template instantiation of length
  template int length(const bvec &v);

  //! Template instantiation of sum
  template double sum(const Vec<double> &v);
  //! Template instantiation of sum
  template complex<double> sum(const cvec &v);
  //! Template instantiation of sum
  template short sum(const svec &v);
  //! Template instantiation of sum
  template int sum(const ivec &v);
  //! Template instantiation of sum
  template bin sum(const bvec &v);

  //! Template instantiation of sum_sqr
  template double sum_sqr(const vec &v);
  //! Template instantiation of sum_sqr
  template complex<double> sum_sqr(const cvec &v);
  //! Template instantiation of sum_sqr
  template short sum_sqr(const svec &v);
  //! Template instantiation of sum_sqr
  template int sum_sqr(const ivec &v);
  //! Template instantiation of sum_sqr
  template bin sum_sqr(const bvec &v);

  //! Template instantiation of max
  template double max(const vec &in);
  //! Template instantiation of max
  template short max(const svec &in);
  //! Template instantiation of max
  template int max(const ivec &in);

  //! Template instantiation of min
  template double min(const vec &in);
  //! Template instantiation of min
  template short min(const svec &in);
  //! Template instantiation of min
  template int min(const ivec &in);

  //! Template instantiation of max_index
  template int max_index(const vec &in);
  //! Template instantiation of max_index
  template int max_index(const svec &in);
  //! Template instantiation of max_index
  template int max_index(const ivec &in);

  //! Template instantiation of min_index
  template int min_index(const vec &in);
  //! Template instantiation of min_index
  template int min_index(const svec &in);
  //! Template instantiation of min_index
  template int min_index(const ivec &in);

  //! Template instantiation of product
  template double product(const vec &v);
  //! Template instantiation of product
  template short product(const svec &v);
  //! Template instantiation of product
  template int product(const ivec &v);

  //! Template instantiation of cross
  template vec cross(const vec &v1, const vec &v2);
  //! Template instantiation of cross
  template ivec cross(const ivec &v1, const ivec &v2);
  //! Template instantiation of cross
  template svec cross(const svec &v1, const svec &v2);

  //! Template instantiation of geometric_mean
  template double geometric_mean(const vec &v);
  //! Template instantiation of geometric_mean
  template double geometric_mean(const svec &v);
  //! Template instantiation of geometric_mean
  template double geometric_mean(const ivec &v);

  //! Template instantiation of median
  template double median(const vec &v);
  //! Template instantiation of median
  template double median(const svec &v);
  //! Template instantiation of median
  template double median(const ivec &v);

  //! Template instantiation of norm
  template double norm(const  vec &v);
  // Template instantiation of norm
  template double norm(const  svec &v);
  //! Template instantiation of norm
  template double norm(const  ivec &v);

  //! Template instantiation of norm
  template double norm(const  vec &v, int p);
  //! Template instantiation of norm
  template double norm(const  svec &v, int p);
  //! Template instantiation of norm
  template double norm(const  ivec &v, int p);

  //! Template instantiation of variance
  template double variance(const  vec &v);
  //! Template instantiation of variance
  template double variance(const  svec &v);
  //! Template instantiation of variance
  template double variance(const  ivec &v);

  //! Template instantiation of energy
  template double energy(const  vec &v);
  //! Template instantiation of energy
  template double energy(const cvec &v);
  //! Template instantiation of energy
  template double energy(const  svec &v);
  //! Template instantiation of energy
  template double energy(const  ivec &v);

  //! Template instantiation of reverse
  template vec reverse(const vec &in);
  //! Template instantiation of reverse
  template cvec reverse(const cvec &in);
  //! Template instantiation of reverse
  template svec reverse(const svec &in);
  //! Template instantiation of reverse
  template ivec reverse(const ivec &in);
  //! Template instantiation of reverse
  template bvec reverse(const bvec &in);

  //! Template instantiation of repeat
  template vec repeat(const vec &v, int norepeats);
  //! Template instantiation of repeat
  template cvec repeat(const cvec &v, int norepeats);
  //! Template instantiation of repeat
  template svec repeat(const svec &v, int norepeats);
  //! Template instantiation of repeat
  template ivec repeat(const ivec &v, int norepeats);
  //! Template instantiation of repeat
  template bvec repeat(const bvec &v, int norepeats);

  //! Template instantiation of apply_function
  template vec apply_function(float (*f)(float), const vec &data);
  //! Template instantiation of apply_function
  template vec apply_function(double (*f)(double), const vec &data);
  //! Template instantiation of apply_function
  template cvec apply_function(complex<double> (*f)(complex<double>), const cvec &data);
  //! Template instantiation of apply_function
  template svec apply_function(short (*f)(short), const svec &data);
  //! Template instantiation of apply_function
  template ivec apply_function(int (*f)(int), const ivec &data);
  //! Template instantiation of apply_function
  template bvec apply_function(bin (*f)(bin), const bvec &data);

  //! Template instantiation of sort
  template void sort(vec &data);
  //! Template instantiation of sort
  template void sort(svec &data);
  //! Template instantiation of sort
  template void sort(ivec &data);
  //! Template instantiation of sort
  template void sort(bvec &data);

  //! Template instantiation of sort_index
  template ivec sort_index(const vec &data);
  //! Template instantiation of sort_index
  template ivec sort_index(const svec &data);
  //! Template instantiation of sort_index
  template ivec sort_index(const ivec &data);
  //! Template instantiation of sort_index
  template ivec sort_index(const bvec &data);

  //! Template instantiation of zero_pad
  template ivec zero_pad(const ivec &v, int n);
  //! Template instantiation of zero_pad
  template vec zero_pad(const vec &v, int n);
  //! Template instantiation of zero_pad
  template cvec zero_pad(const cvec &v, int n);
  //! Template instantiation of zero_pad
  template bvec zero_pad(const bvec &v, int n);

  //! Template instantiation of zero_pad
  template ivec zero_pad(const ivec &v);
  //! Template instantiation of zero_pad
  template vec zero_pad(const vec &v);
  //! Template instantiation of zero_pad
  template cvec zero_pad(const cvec &v);
  //! Template instantiation of zero_pad
  template bvec zero_pad(const bvec &v);

  //! Template instantiation of zero_pad
  template mat   zero_pad(const mat &, int, int);
  //! Template instantiation of zero_pad
  template cmat  zero_pad(const cmat &, int, int);
  //! Template instantiation of zero_pad
  template imat  zero_pad(const imat &, int, int);
  //! Template instantiation of zero_pad
  template bmat  zero_pad(const bmat &, int, int);

  //! Template instantiation of sum
  template double sum(const mat &m);
  //! Template instantiation of sum
  template complex<double> sum(const cmat &m);
  //! Template instantiation of sum
  template short sum(const smat &m);
  //! Template instantiation of sum
  template int sum(const imat &m);
  //! Template instantiation of sum
  template bin sum(const bmat &m);

  //! Template instantiation of sum_sqr
  template double sum_sqr(const mat &m);
  //! Template instantiation of sum_sqr
  template complex<double> sum_sqr(const cmat &m);
  //! Template instantiation of sum_sqr
  template short sum_sqr(const smat &m);
  //! Template instantiation of sum_sqr
  template int sum_sqr(const imat &m);
  //! Template instantiation of sum_sqr
  template bin sum_sqr(const bmat &m);

  //! Template instantiation of max
  template double max(const mat &m);
  //! Template instantiation of max
  template int max(const imat &m);

  //! Template instantiation of min
  template double min(const mat &m);
  //! Template instantiation of min
  template int min(const imat &m);

  //! Template instantiation of max_index
  template void max_index(const mat &m, int &row, int &col);
  //! Template instantiation of max_index
  template void max_index(const imat &m, int &row, int &col);

  //! Template instantiation of min_index
  template void min_index(const mat &m, int &row, int &col);
  //! Template instantiation of min_index
  template void min_index(const imat &m, int &row, int &col);

  //! Template instantiation of product
  template double product(const mat &m);
  // Template instantiation of product
  //template complex<double> product(const cmat &v);
  //! Template instantiation of product
  template short product(const smat &m);
  //! Template instantiation of product
  template int product(const imat &m);

  //! Template instantiation of diag
  template vec diag(const mat &in);
  //! Template instantiation of diag
  template cvec diag(const cmat &in);

  //! Template instantiation of diag
  template void diag(const vec &in, mat &m);
  //! Template instantiation of diag
  template void diag(const cvec &in, cmat &m);

  //! Template instantiation of diag
  template mat diag(const vec &v);
  //! Template instantiation of diag
  template cmat diag(const cvec &v);

  //! Template instantiation of bidiag
  template mat bidiag(const vec &, const vec &);
  //! Template instantiation of bidiag
  template cmat bidiag(const cvec &, const cvec &);

  //! Template instantiation of bidiag
  template void bidiag(const vec &, const vec &, mat &);
  //! Template instantiation of bidiag
  template void bidiag(const cvec &, const cvec &, cmat &);

  //! Template instantiation of bidiag
  template void bidiag(const mat &, vec &, vec &);
  //! Template instantiation of bidiag
  template void bidiag(const cmat &, cvec &, cvec &);

  //! Template instantiation of tridiag
  template mat tridiag(const vec &main, const vec &, const vec &);
  //! Template instantiation of tridiag
  template cmat tridiag(const cvec &main, const cvec &, const cvec &);

  //! Template instantiation of tridiag
  template void tridiag(const vec &main, const vec &, const vec &, mat &);
  //! Template instantiation of tridiag
  template void tridiag(const cvec &main, const cvec &, const cvec &, cmat &);

  //! Template instantiation of tridiag
  template void tridiag(const mat &m, vec &, vec &, vec &);
  //! Template instantiation of tridiag
  template void tridiag(const cmat &m, cvec &, cvec &, cvec &);

  //! Template instantiation of trace
  template double trace(const mat &in);
  //! Template instantiation of trace
  template complex<double> trace(const cmat &in);
  //! Template instantiation of trace
  template short trace(const smat &in);
  //! Template instantiation of trace
  template int trace(const imat &in);
  //! Template instantiation of trace
  template bin trace(const bmat &in);

  //! Template instantiation of transpose
  template void transpose(const mat &m, mat &out);
  //! Template instantiation of transpose
  template void transpose(const cmat &m, cmat &out);
  //! Template instantiation of transpose
  template void transpose(const smat &m, smat &out);
  //! Template instantiation of transpose
  template void transpose(const imat &m, imat &out);
  //! Template instantiation of transpose
  template void transpose(const bmat &m, bmat &out);

  //! Template instantiation of transpose
  template mat transpose(const mat &m);
  //! Template instantiation of transpose
  template cmat transpose(const cmat &m);
  //! Template instantiation of transpose
  template smat transpose(const smat &m);
  //! Template instantiation of transpose
  template imat transpose(const imat &m);
  //! Template instantiation of transpose
  template bmat transpose(const bmat &m);

  //! Template instantiation of repeat
  template mat repeat(const mat &m, int norepeats);
  //! Template instantiation of repeat
  template cmat repeat(const cmat &m, int norepeats);
  //! Template instantiation of repeat
  template smat repeat(const smat &m, int norepeats);
  //! Template instantiation of repeat
  template imat repeat(const imat &m, int norepeats);
  //! Template instantiation of repeat
  template bmat repeat(const bmat &m, int norepeats);

  //! Template instantiation of apply_function
  template mat apply_function(float (*f)(float), const mat &data);
  //! Template instantiation of apply_function
  template mat apply_function(double (*f)(double), const mat &data);
  //! Template instantiation of apply_function
  template cmat apply_function(complex<double> (*f)(complex<double>), const cmat &data);
  //! Template instantiation of apply_function
  template smat apply_function(short (*f)(short), const smat &data);
  //! Template instantiation of apply_function
  template imat apply_function(int (*f)(int), const imat &data);
  //! Template instantiation of apply_function
  template bmat apply_function(bin (*f)(bin), const bmat &data);

  //! Template instantiation of rvectorize
  template  vec rvectorize(const  mat &m);
  //! Template instantiation of rvectorize
  template cvec rvectorize(const cmat &m);
  //! Template instantiation of rvectorize
  template  ivec rvectorize(const  imat &m);
  //! Template instantiation of rvectorize
  template  bvec rvectorize(const  bmat &m);

  //! Template instantiation of cvectorize
  template  vec cvectorize(const  mat &m);
  //! Template instantiation of cvectorize
  template cvec cvectorize(const cmat &m);
  //! Template instantiation of cvectorize
  template  ivec cvectorize(const  imat &m);
  //! Template instantiation of cvectorize
  template  bvec cvectorize(const  bmat &m);

  //! Template instantiation of reshape
  template  mat reshape(const  mat &m, int rows, int cols);
  //! Template instantiation of reshape
  template cmat reshape(const cmat &m, int rows, int cols);
  //! Template instantiation of reshape
  template  imat reshape(const  imat &m, int rows, int cols);
  //! Template instantiation of reshape
  template  bmat reshape(const  bmat &m, int rows, int cols);

  //! Template instantiation of reshape
  template  mat reshape(const  vec &m, int rows, int cols);
  //! Template instantiation of reshape
  template cmat reshape(const cvec &m, int rows, int cols);
  //! Template instantiation of reshape
  template  imat reshape(const  ivec &m, int rows, int cols);
  //! Template instantiation of reshape
  template  bmat reshape(const  bvec &m, int rows, int cols);

  //! Template instantiation of upsample
  template vec upsample(const vec &v, int usf);
  //! Template instantiation of upsample
  template cvec upsample(const cvec &v, int usf);
  //! Template instantiation of upsample
  template svec upsample(const svec &v, int usf);
  //! Template instantiation of upsample
  template ivec upsample(const ivec &v, int usf);
  //! Template instantiation of upsample
  template bvec upsample(const bvec &v, int usf);

  //! Template instantiation of upsample
  template mat upsample(const mat &v, int usf);
  //! Template instantiation of upsample
  template cmat upsample(const cmat &v, int usf);
  //! Template instantiation of upsample
  template smat upsample(const smat &v, int usf);
  //! Template instantiation of upsample
  template imat upsample(const imat &v, int usf);
  //! Template instantiation of upsample
  template bmat upsample(const bmat &v, int usf);

  //! Template instantiation of upsample
  template void upsample(const vec &v, int usf,  vec & u);
  //! Template instantiation of upsample
  template void upsample(const cvec &v, int usf,  cvec & u);
  //! Template instantiation of upsample
  template void upsample(const svec &v, int usf,  svec & u);
  //! Template instantiation of upsample
  template void upsample(const ivec &v, int usf,  ivec & u);
  //! Template instantiation of upsample
  template void upsample(const bvec &v, int usf,  bvec & u);

  //! Template instantiation of upsample
  template void upsample(const mat &v, int usf,  mat & u);
  //! Template instantiation of upsample
  template void upsample(const cmat &v, int usf,  cmat & u);
  //! Template instantiation of upsample
  template void upsample(const smat &v, int usf,  smat & u);
  //! Template instantiation of upsample
  template void upsample(const imat &v, int usf,  imat & u);
  //! Template instantiation of upsample
  template void upsample(const bmat &v, int usf,  bmat & u);

  //! Template instantiation of liniterp
  template vec lininterp(const vec &v, int usf);
  //! Template instantiation of liniterp
  template cvec lininterp(const cvec &v, int usf);

  //! Template instantiation of liniterp
  template mat lininterp(const mat &v, int usf);
  //! Template instantiation of liniterp
  template cmat lininterp(const cmat &v, int usf);

  //! Template instantiation of liniterp
  template void lininterp(const vec &v, int usf,  vec & u);
  //! Template instantiation of liniterp
  template void lininterp(const cvec &v, int usf,  cvec & u);

  //! Template instantiation of liniterp
  template void lininterp(const mat &v, int usf,  mat & u);
  //! Template instantiation of liniterp
  template void lininterp(const cmat &v, int usf,  cmat & u);

} //namespace itpp
